#!/bin/bash


########################################################################
# Script for paper titled 'mbonsai Application Package for Sequence Classication
#
# The following scripts replicates mbonsai commands executed at the UNIX/LINUX shell or Terminal app on Mac OS X.
#
# To run this script:
# 1. Create parent directory (in the tutorial the directory is named as parentdir) and save the zip archive inside.
# $ mkdir parentdir
# $ cd parentdir 
# ~/parentdir$ 
#
# 2. Unzip mbonsai application in parentdir. The directory should contain 2 files and 2 directories. Then go to mbonsai/cmd directory, and install the program
# ~/parentdir$ unzip mbonsai.zip
# ~/parentdir$ ls
# data  license.txt  mbonsai  mbonsaiscript.sh  
# ~/parentdir$ cd mbonsai/cmd
# ~/parentdir/mbonsai/cmd$ make
# # Installation of mbonsai can be checked with test script
# ~/parentdir$ cd mbonsai/check
# ~/parentdir/mbonsai/check$ bash test.sh
#
#
# 3. Go back to main parent directory, type the following at the command prompt to execute all commands for the tutoral:
#  ~/parentdir$ bash mbonsaiscript.sh
#
# 4. To execute individual commands, at the ~/parentdir$ command prompt, type the corresponding commands at the command prompt at the parent folder. e.g.  
# mbonsai/cmd/mbonsai p=seq c=cls i=data/input.csv O=result
#
# The output and classification tree models are stored in new folders defined at O= parameter as result_figure1, result_figure1_index, result_core, and result_brand directories.
#
# Note 1: seed= parameter can be used to set the random seed in order to replicate the same results as shown in the paper in an OS X 10.X environment.  Results may differ slightly depending on the random number generator in your system.
# Note 2: Users can modify the color of the graph by editing the colorSet in dictColor function in \code{mbonsai/view/mdtree.rb}. 
########################################################################
# 
# Now, Let's proceed with trying the commands in the tutorial. First, go to the parent directory indicated as follows 
# ~/parentdir$

# Clear old results files before rerunning the script or tutorial commands
rm -rf result_figure1
rm -rf result_figure1_index
rm -rf result_core
rm -rf result_brand

########################################################################
# Table 1 - Input data: figure1/input.csv
# This command shows the diffference of using index and without using index as shown in Figure 1 is generated. 
# Script
## Figure 1
# a) Original decision tree
mbonsai/cmd/mbonsai p=seq c=cls i=data/input.csv O=result_figure1
# -> Result: result_figure1/model.txt

# b) Decision tree with alphabet index. Set random seed at 10
mbonsai/cmd/mbonsai p=seq:2 c=cls i=data/input.csv O=result_figure1_index seed=10
# -> Result: result_figure1_index/model.txt
#
# Pg 12 - Shows example parameter to execute the command, which is equivalent to Figure 1a command. 
# 
########################################################################
# Table 7: Input data: core/core.csv, core/core_test.csv

# Script
## Application I: Identification of core customers
# Core Customer classification model (pg 12)
# Use random seed at 100
mbonsai/cmd/mbonsai i=data/core.csv p=ProfitPattern:2::true,VisitPattern n=Profit,Visit c=Target O=result_core seed=100
# -> Result: core/result/model.txt

# Prediction using classification model (pg 14)
mbonsai/cmd/mbonsai -predict i=data/core_test.csv o=result_core/predict.csv I=result_core
# -> Result: result_core/predict.csv

# Calculation results of consufion matrix from predict.csv is created by statistical formulas which is not part of mbonsai. The results can be derived by executing the following ruby script

mbonsai/cmd/mcm i=result_core/predict.csv ac=Target pc=predict O=result_core/evalAcc
# -> Result: result_core/evalAcc directory

# Figure 3: Visualization of the Decision tree (pg 15)

mbonsai/view/mdtree.rb i=result_core/model.pmml o=result_core/tree.html
# -> Result: core/result/tree.csv

mbonsai/view/mdtree.rb i=result_core/model.pmml o=result_core/tree_bar.html -bar
# -> Result: core/result/tree.csv

# Refine the model (pg 15)

mbonsai/cmd/mbonsai i=data/core.csv p=ProfitPattern:2::true,VisitPattern n=Profit,Visit c=Target seed=100 alpha=0.00001 leafSize=1500 O=result_core_mod
# -> Result: result_core_mod/model.txt

# Figure 4: Visualize the refined model with alpha set as 0.00001 with minimum 1500 samples in each leaf (pg 17)

mbonsai/view/mdtree.rb i=result_core_mod/model.pmml o=result_core_mod/tree.html -bar
# -> Result: result_core_mod/tree.html

########################################################################
# Table 8: Input data: brand/input.csv
# Script
## Application II: Prediction of Diaper Purchase
# Create a decision tree model (pg 18)

mbonsai/cmd/mbonsai i=data/brand.csv p=BrandPattern:2 c=Target O=result_brand cv=5 seed=500
# -> Result: result_brand/model.txt

# Figure 5: Decision tree for brand pattern dataset (pg 20)

mbonsai/view/mdtree.rb i=result_brand/model.pmml o=result_brand/tree.html -bar
# -> Result: result_brand/tree.html

# Calculation results of consufion matrix from predict.csv is created by statistical formulas which is not part of mbonsai. The results can be derived by executing the following ruby script (pg 20)

mbonsai/cmd/mcm i=result_brand/predict_1se.csv ac=Target pc=predict O=result_brand/evalAcc_1se
# -> Result: result_brand/evalAcc_1se directory

mbonsai/cmd/mcm i=result_brand/predict_min.csv ac=Target pc=predict O=result_brand/evalAcc_min
# -> Result: result_brand/evalAcc_min directory



